'''
(c) 2010-2012 Thomas Holder (https://github.com/speleo3/pymol-psico)
(c) 2016 Thomas Holder, Schrodinger, Inc.

License: BSD-2-Clause
'''

from pymol import cmd, CmdException


def pdb2pqr(name, selection='all', ff='amber', debump=1, opt=1, assignonly=0,
        ffout='', ph=None, neutraln=0, neutralc=0, state=-1, preserve=0,
        exe='pdb2pqr', quiet=1):
    '''
DESCRIPTION

    Creates a new molecule object from a selection and adds missing atoms,
    assignes charges and radii using PDB2PQR.

    http://www.poissonboltzmann.org/pdb2pqr/

USAGE

    pdb2pqr name [, selection [, ff [, debump [, opt [, assignonly [, ffout [,
        ph [, neutraln [, neutralc [, state [, preserve ]]]]]]]]]]]

ARGUMENTS

    name = string: name of object to create or modify

    selection = string: atoms to include in the new object {default: all}

    ff = string: forcefield {default: amber}
    '''
    debump, opt, assignonly = int(debump), int(opt), int(assignonly)
    neutraln, neutralc = int(neutraln), int(neutralc)
    quiet = int(quiet)

    args = ['--ff=' + ff, '--chain']
    if not debump:
        args.append('--nodebump')
    if not opt:
        args.append('--noopt')
    if assignonly:
        args.append('--assign-only')
    if ffout:
        args.append('--ffout=' + ffout)
    if ph is not None:
        args.append('--with-ph=%f' % float(ph))
    if neutraln:
        args.append('--neutraln')
    if neutralc:
        args.append('--neutralc')
    if not quiet:
        args.append('--verbose')

    r = pdb2pqr_cli(name, selection, args, state, preserve, exe, quiet)

    if not quiet:
        if r:
            print(r)

        print(' pdb2pqr: done')

    return r


def pdb2pqr_cli(name, selection, options, state=-1, preserve=0,
        exe='pdb2pqr', quiet=1, fixrna=0, _proclist=None):
    import os, tempfile, subprocess, shutil

    state, preserve, quiet = int(state), int(preserve), int(quiet)

    if cmd.is_string(options):
        import shlex
        options = shlex.split(options)

    args = [cmd.exp_path(exe)] + list(options)

    tmpdir = tempfile.mkdtemp()
    # Input format should be PDB, but use PQR instead because we can support
    # multi-state assemblies by not writing MODEL records.
    infile = os.path.join(tmpdir, 'in.pqr')
    outfile = os.path.join(tmpdir, 'out.pqr')
    args.extend([infile, outfile])

    # For some reason, catching stdout/stderr with PIPE and communicate()
    # blocks terminate() calls from terminating early. Using a file
    # redirect doesn't show this problem.
    f_stdout = open(os.path.join(tmpdir, 'stdout.txt'), 'w+')

    # RNA resdiue names must be RA, RC, RG, and RU
    tmpmodel = ''
    if int(fixrna) and cmd.count_atoms('(%s) & resn A+C+G+U' % (selection)):
        tmpmodel = cmd.get_unused_name('_tmp')
        cmd.create(tmpmodel, selection, zoom=0)
        cmd.alter(tmpmodel + ' & polymer.nucleic & resn A+C+G+U',
                'resn = "R" + resn')
        selection = tmpmodel

    try:
        cmd.save(infile, selection, state)

        p = subprocess.Popen(args, cwd=tmpdir,
                stdin=subprocess.PIPE,  # Windows pythonw fix
                stdout=f_stdout,
                stderr=f_stdout)
        p.stdin.close()  # Windows pythonw fix

        if _proclist is not None:
            _proclist.append(p)

        p.wait()

        # This allows PyMOL to capture it and display the output in the GUI.
        f_stdout.seek(0)
        print(f_stdout.read().rstrip())

        if p.returncode == -15: # SIGTERM
            raise CmdException('pdb2pqr terminated')

        if p.returncode != 0:
            raise CmdException('%s failed with exit status %d' %
                    (args[0], p.returncode))

        warnings = [L[10:] for L in open(outfile) if L.startswith('REMARK   5')]
        warnings = ''.join(warnings).strip()

        cmd.load(outfile, name)

        return warnings

    except OSError as e:
        print(e)
        raise CmdException('Cannot execute "%s"' % (exe))
    finally:
        if tmpmodel:
            cmd.delete(tmpmodel)

        f_stdout.close()

        if not preserve:
            shutil.rmtree(tmpdir)
        elif not quiet:
            print(' Notice: not deleting', tmpdir)


def _is_exe(exe):
    import os, sys
    if os.path.exists(exe):
        return True
    if sys.platform.startswith('win'):
        return os.path.exists(exe + '.exe')
    return False


def prepwizard(name, selection='all', options='', state=-1,
        preserve=0, exe='$SCHRODINGER/utilities/prepwizard', quiet=1,
        _proclist=None):
    '''
DESCRIPTION

    Run the SCHRODINGER Protein Preparation Wizard. Builds missing side
    chains and converts MSE to MET. Other non-default options need to be
    passed with the "options=" argument.

USAGE

    prepwizard name [, selection [, options [, state ]]]

ARGUMENTS

    name = str: name of object to create

    selection = str: atoms to send to the wizard {default: all}

    options = str: additional command line options for prepwizard

    state = int: object state {default: -1 (current)}
    '''
    import os, tempfile, subprocess, shutil, shlex

    state, preserve, quiet = int(state), int(preserve), int(quiet)

    exe = cmd.exp_path(exe)
    if not _is_exe(exe):
        if 'SCHRODINGER' not in os.environ:
            print(' Warning: SCHRODINGER environment variable not set')
        raise CmdException('no such script: ' + exe)

    args = [exe, '-mse', '-fillsidechains', '-WAIT']

    if options:
        if cmd.is_string(options):
            options = shlex.split(options)
        args.extend(options)

    tmpdir = tempfile.mkdtemp()
    infile = 'in.pdb'
    outfile = 'out.mae'
    args.extend([infile, outfile])

    try:
        cmd.save(os.path.join(tmpdir, infile), selection, state)

        env = dict(os.environ)
        env.pop('PYTHONEXECUTABLE', '')  # messes up Python on Mac

        import pymol
        if pymol.IS_WINDOWS:
            # Fix for 2020-4 (PYMOL-3572)
            import ctypes
            ctypes.windll.kernel32.SetDllDirectoryW(None)

        p = subprocess.Popen(args, cwd=tmpdir,
                env=env,
                universal_newlines=True,
                stdin=subprocess.PIPE,  # Windows pythonw fix
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT)

        if _proclist is not None:
            _proclist.append(p)

        print(p.communicate()[0].rstrip())

        if p.wait() == -15: # SIGTERM
            raise CmdException('prepwizard terminated')

        if p.returncode != 0:
            raise CmdException('%s failed with exit status %d' % (args[0], p.returncode))

        cmd.load(os.path.join(tmpdir, outfile), name)
    except OSError as e:
        print(e)
        raise CmdException('Cannot execute "%s"' % (exe))
    finally:
        logfile = os.path.join(tmpdir, 'in.log')
        if os.path.exists(logfile):
            with open(logfile) as handle:
                print(handle.read())
        if not preserve:
            shutil.rmtree(tmpdir)
        elif not quiet:
            print(' Notice: not deleting', tmpdir)

    if not quiet:
        print(' prepwizard: done')


# vi: ts=4:sw=4:smarttab:expandtab



from enum import IntEnum

from pymol import setting
from pymol import save_shortcut as shortcut_saver
from pymol.keyboard import get_default_keys
from pymol.shortcut_dict import shortcut_dict_ref

class ShortcutIndex(IntEnum):
    COMMAND = 0
    DESCRIPT = 1
    USER_DEF = 2

class ShortcutManager():
    def __init__(self, saved_shortcuts, cmd):
        self.cmd = cmd
        self.saved_shortcuts = saved_shortcuts
        self.default_bindings = get_default_keys(self.cmd)
        self.cmd.shortcut_dict = {key: list(value) for key,value in shortcut_dict_ref.items()}

        # Tuple of keys that are reserved for the system
        self.reserved_keys = ('CTRL-S','CTRL-E','CTRL-O','CTRL-M','up','down')

    def check_saved_dict(self):
        '''
        Checks if there is a saved file. Updates shortcut_dict if needed.
        '''
        for key, saved_shortcut_list in self.saved_shortcuts.items():
            if key in self.cmd.shortcut_dict:
                self.cmd.shortcut_dict[key][ShortcutIndex.USER_DEF] = saved_shortcut_list[ShortcutIndex.USER_DEF]
            else:
                self.cmd.shortcut_dict.update({key:saved_shortcut_list})

    def check_key_mappings(self):
        '''
        Corrects mismatched key mappings and adds missing keys.
        There are three locations of shortcuts:
        1) cmd.key_mappings: Current state of key mappings for PyMOL
        2) cmd.default_bindings: Default values of key mappings
        3) cmd.shortcut_dict: Belongs to this class instance, for table
        '''
        current_mappings = self.cmd.key_mappings
        missing_keys = []
        mismatch_keys = []
        empty_keys = []
        manual_reset_keys = []

        for key in current_mappings:
            if not current_mappings[key]:
                empty_keys.append(key)
            if key in self.default_bindings:
                if current_mappings[key] != self.default_bindings[key]:
                    mismatch_keys.append(key)
                elif current_mappings[key] == self.default_bindings[key] and self.cmd.shortcut_dict[key][2]:
                    manual_reset_keys.append(key)
            else:
                missing_keys.append(key)
        for key in mismatch_keys:
            self.cmd.shortcut_dict[key][ShortcutIndex.USER_DEF] = str(current_mappings[key])
        for key in manual_reset_keys:
            self.cmd.shortcut_dict[key][ShortcutIndex.USER_DEF] = ''
        for key in missing_keys:
            if key in self.cmd.shortcut_dict:
                self.cmd.shortcut_dict[key][ShortcutIndex.USER_DEF] = str(current_mappings[key])
            else:
                # User defined keys using set_key outside of menu
                self.cmd.shortcut_dict[key] = ['','',str(current_mappings[key])]
        for key in empty_keys:
            self.cmd.shortcut_dict[key][ShortcutIndex.USER_DEF] = 'Deleted'

    def save_shortcuts(self):
        '''
        Removes unused keys before calling save_shortcuts.
        These unused keys represent keybinding defined by the user but then assigned to nothing.
        '''
        self.remove_unused(self.cmd.shortcut_dict)
        shortcut_saver.save_shortcuts(self.cmd.shortcut_dict)

    def remove_unused(self, full_dict):
        '''
        Removes items from dictionary if first and third element are missing from the value list. 
        '''
        unused_keys = []
        for key in full_dict:
            if not full_dict[key][0] and not full_dict[key][2]:
                unused_keys.append(key)
        for key in unused_keys:
            del full_dict[key]
        return full_dict

    def reset_all_default(self):
        '''
        Iterates over all values to restore their default commands.
        This will restore them to the values from keyboard.py
        '''
        delete_keys = []

        for key in self.cmd.shortcut_dict:
            self.cmd.shortcut_dict[key][ShortcutIndex.USER_DEF] = ''

            if key in self.default_bindings:
                binding = self.default_bindings[key]
                if isinstance(binding,str):
                    self.cmd.set_key(key,binding)
                elif isinstance(binding,tuple):
                    self.cmd.set_key(key,binding[0],binding[1],binding[2])
                else:
                    print("Incorrect type found when resetting defaults")
            else:
                self.cmd.set_key(key,'')
                delete_keys.append(key)

        for key in delete_keys:
            del self.cmd.shortcut_dict[key]
            del self.cmd.key_mappings[key]

        print("Restored default keybindings")

    def create_new_shortcut(self, new_key, new_binding):
        '''
        Creates a new shortcut after checking existing and reserved keys.
        '''
        if new_key not in self.cmd.shortcut_dict and new_key not in self.reserved_keys:
            try:
                self.cmd.set_key(new_key,new_binding)
                print('Assigning ',new_key, ' to ',new_binding)
            except Exception:
                print("This cannot be bound.")
            else:
                self.cmd.shortcut_dict.update({new_key:['','',new_binding]})
        elif new_key in self.cmd.shortcut_dict:
            try:
                self.cmd.set_key(new_key,new_binding)
                print('Assigning ',new_key, ' to ',new_binding)
            except Exception:
                print("This cannot be bound.")
            else:
                self.cmd.shortcut_dict[new_key][2] = new_binding
        else:
            print("This key is reserved.")



'''
PyMOL APBS GUI Plugin

(c) Schrodinger, Inc.
'''

import sys
import os
import importlib

from pymol.Qt import QtCore, QtWidgets
from pymol.Qt.utils import loadUi, AsyncFunc, MainThreadCaller

getOpenFileNames = QtWidgets.QFileDialog.getOpenFileNames

from . import electrostatics
from .qtwidgets import ResizableMessageBox as QMessageBox


class SilentAbort(Exception):
    pass


class StdOutCapture:
    '''
Redirect stdout and/or stderr to a temporary file until 'release' is called.
    '''

    def __init__(self, stdout=True, stderr=True):
        import tempfile
        self.temp = tempfile.TemporaryFile('w+')
        self.save_stdout = None
        self.save_stderr = None
        if stdout:
            self.save_stdout = sys.stdout
            sys.stdout = self.temp
        if stderr:
            self.save_stderr = sys.stderr
            sys.stderr = self.temp

    def release(self):
        if self.save_stdout is not None:
            sys.stdout = self.save_stdout
        if self.save_stderr is not None:
            sys.stderr = self.save_stderr
        self.temp.seek(0)
        content = self.temp.read()
        self.temp.close()
        return content


def run_impl(form, _self):
    '''
Execute the pipeline (prep, apbs, surface vis)
    '''
    selection = form.input_sele.currentText()
    prep_name = ''
    map_name = ''
    ramp_name = ''
    group_name = ['']

    def get_name(lineinput):
        name = lineinput.text()
        if not name:
            name = _self.get_unused_name(lineinput.placeholderText())
        return name

    def do_group(name):
        if form.check_no_group.isChecked():
            return
        if not group_name[0]:
            group_name[0] = _self.get_unused_name('run', 1)
        _self.group(group_name[0], name)

    if form.do_prepare.isChecked():
        from . import creating as pc

        prep_name = get_name(form.prep_name)
        method = form.prep_method.currentText()
        warnings = ''

        get_res_charge = lambda resn: ((-1) if resn in ('GLU', 'ASP') else
                (1) if resn in ('ARG', 'LYS') else 0)

        def get_cb_pseudocharge(resn, name):
            if name != 'CB':
                return 0
            return get_res_charge(resn)

        try:
            state = _self.get_selection_state(selection)
        except BaseException as e:
            state = -1

        if method == 'pdb2pqr':
            warnings = pc.pdb2pqr_cli(prep_name, selection,
                    options=form.pdb2pqr_args.text(),
                    quiet=0,
                    state=state,
                    preserve=form.check_preserve.isChecked(),
                    fixrna=form.pdb2pqr_fixrna.isChecked(),
                    _proclist=form._proclist,
                    exe=form.pdb2pqr_exe.text())
            if form.pdb2pqr_ignore_warnings.isChecked():
                warnings = ''
        elif method == 'protein_assign_charges_and_radii':
            _self.create(prep_name, selection)
            _self.util.protein_assign_charges_and_radii(prep_name)
        elif method.startswith('prepwizard'):
            pc.prepwizard(prep_name, selection,
                    options=form.prepwizard_args.text(),
                    _proclist=form._proclist,
                    quiet=0,
                    preserve=form.check_preserve.isChecked())
            _self.alter(prep_name, 'elec_radius = vdw')
        elif method.startswith('use formal'):
            _self.create(prep_name, selection)
            _self.alter(prep_name,
                        '(elec_radius, partial_charge) = (vdw, formal_charge)')
        elif method.startswith('use CB'):
            _self.create(prep_name, selection)
            _self.alter(
                prep_name,
                '(elec_radius, partial_charge) = (vdw, getpc(resn, name))',
                space={'getpc': get_cb_pseudocharge})
        elif method.startswith('use CA'):
            _self.create(prep_name, '(%s) & name CA' % (selection))
            _self.alter(
                prep_name,
                '(elec_radius, vdw, partial_charge) = (3.0, 3.0, getpc(resn))',
                space={'getpc': get_res_charge})
        elif method.startswith('use vdw'):
            _self.create(prep_name, selection)
            _self.alter(prep_name, 'elec_radius = vdw')
        else:
            raise ValueError('unknown method: ' + method)

        selection = prep_name
        form.input_sele.addItem(prep_name)
        do_group(prep_name)

        if warnings:
            @form._callInMainThread
            def result():
                msgbox = QMessageBox(QMessageBox.Question, 'Continue?',
                    method + ' emmitted warnings, do you want to continue?',
                    QMessageBox.Yes | QMessageBox.No , form._dialog)
                msgbox.setDetailedText(warnings)
                return msgbox.exec_()

            if result == QMessageBox.No:
                raise SilentAbort

    if form.do_apbs.isChecked():
        map_name = get_name(form.apbs_map)
        template = form.apbs_template.toPlainText()
        electrostatics.map_new_apbs(
            map_name,
            selection,
            grid=form.apbs_grid.value(),
            focus=form.focus_sele.text(),
            quiet=0,
            preserve=form.check_preserve.isChecked(),
            exe=form.apbs_exe.text(),
            _template=template,
            _proclist=form._proclist,
            assign=0)
        form.surf_map.addItem(map_name)
        do_group(map_name)

    if form.do_surface.isChecked():
        if not map_name:
            map_name = form.surf_map.currentText()
            if not map_name:
                raise ValueError('missing map')

        if not prep_name:
            prep_name = _self.get_object_list(selection)[0]

        ramp_name = get_name(form.surf_ramp)

        v = form.surf_range.value()
        _self.ramp_new(ramp_name, map_name, [-v, 0, v])
        do_group(ramp_name)

        sas = 'Accessible' in form.surf_type.currentText()
        _self.set('surface_ramp_above_mode', not sas, prep_name)
        _self.set('surface_solvent', sas, prep_name)
        _self.set('surface_color', ramp_name, prep_name)
        _self.show('surface', selection)


def dialog(_self=None):
    if _self is None:
        from pymol import cmd as _self

    dialog = QtWidgets.QDialog()
    uifile = os.path.join(os.path.dirname(__file__), 'apbs.ui')
    form = loadUi(uifile, dialog)
    form._dialog = dialog
    form._proclist = []

    def set_apbs_in(contents):
        form.apbs_template.setPlainText(contents.strip())

    # hide options widgets
    form.optarea_prep.setVisible(False)
    form.optarea_apbs.setVisible(False)
    form.optarea_surf.setVisible(False)
    form.optarea_other.setVisible(False)

    # pre-fill form with likely data
    names = _self.get_object_list()
    names += ['(' + n + ')' for n in _self.get_names('public_selections')]
    if names:
        form.input_sele.clear()
        form.input_sele.addItems([('polymer & ' + name)
                                  if _self.count_atoms('polymer & ' + name) > 0
                                  else name for name in names])
    form.surf_map.addItems(_self.get_names_of_type('object:map'))
    set_apbs_in(electrostatics.template_apbs_in)

    # executables
    from shutil import which
    form.apbs_exe.setText(electrostatics.find_apbs_exe() or 'apbs')
    form.pdb2pqr_exe.setText(
            which('pdb2pqr') or
            which('pdb2pqr30') or
            # acellera::htmd-pdb2pqr provides pdb2pqr_cli
            which('pdb2pqr_cli') or 'pdb2pqr')


    # for async panels
    form._callInMainThread = MainThreadCaller()
    run_impl_async = AsyncFunc(run_impl)

    # "Run" button callback
    def run():
        form.tabWidget.setEnabled(False)
        form.button_ok.clicked.disconnect()
        form.button_ok.clicked.connect(abort)
        form.button_ok.setText('Abort')

        form._capture = StdOutCapture()

        # detach from main thread
        run_impl_async(form, _self)

    # "Run" button "finally" actions (main thread)
    @run_impl_async.finished.connect
    def run_finally(args):
        _, exception = args

        form._proclist[:] = []
        stdout = form._capture.release()
        print(stdout)

        form.button_ok.setText('Run')
        form.button_ok.clicked.disconnect()
        form.button_ok.clicked.connect(run)
        form.button_ok.setEnabled(True)
        form.tabWidget.setEnabled(True)

        if exception is not None:
            handle_exception(exception, stdout)
            return

        quit_msg = "Finished with Success. Close the APBS dialog?"
        if QMessageBox.Yes == QMessageBox.question(
                form._dialog, 'Finished', quit_msg, QMessageBox.Yes,
                QMessageBox.No):
            form._dialog.close()

    def handle_exception(e, stdout):
        if isinstance(e, SilentAbort):
            return

        msg = str(e) or 'unknown error'
        msgbox = QMessageBox(QMessageBox.Critical, 'Error',
                             msg, QMessageBox.Close, form._dialog)
        if stdout.strip():
            msgbox.setDetailedText(stdout)
        msgbox.exec_()

    # "Abort" button callback
    def abort():
        form.button_ok.setEnabled(False)
        while form._proclist:
            p = form._proclist.pop()
            try:
                p.terminate()
                p.returncode = -15  # SIGTERM
            except OSError as e:
                print(e)

    # selection checker
    check_sele_timer = QtCore.QTimer()
    check_sele_timer.setSingleShot(True)

    # grid auto-value
    form.apbs_grid_userchanged = False
    form.apbs_grid.setStyleSheet('background: #ff6')

    @form.apbs_grid.editingFinished.connect
    def _():
        form.apbs_grid_userchanged = True
        form.apbs_grid.setStyleSheet('')

    @check_sele_timer.timeout.connect
    def _():
        has_props = ['no', 'no']

        def callback(partial_charge, elec_radius):
            if partial_charge: has_props[0] = 'YES'
            if elec_radius > 0: has_props[1] = 'YES'

        n = _self.iterate(
            form.input_sele.currentText(),
            'callback(partial_charge, elec_radius)',
            space={'callback': callback})

        # grid auto-value (keep map size in the order of 200x200x200)
        if n > 1 and not form.apbs_grid_userchanged:
            e = _self.get_extent(form.input_sele.currentText())
            volume = (e[1][0] - e[0][0]) * (e[1][1] - e[0][1]) * (e[1][2] - e[0][2])
            grid = max(0.5, volume ** 0.333 / 200.0)
            form.apbs_grid.setValue(grid)

        if n < 1:
            label = 'Selection is invalid'
            color = '#f66'
        elif has_props == ['YES', 'YES']:
            label = 'No preparation necessary, selection has charges and radii'
            form.do_prepare.setChecked(False)
            color = '#6f6'
        else:
            label = 'Selection needs preparation (partial_charge: %s, elec_radius: %s)' % tuple(
                has_props)
            form.do_prepare.setChecked(True)
            color = '#fc6'

        form.label_sele_has.setText(label)
        form.label_sele_has.setStyleSheet('background: %s; padding: 5' % color)

    check_sele_timer.start(0)

    @form.apbs_exe_browse.clicked.connect
    def _():
        fnames = getOpenFileNames(None, filter='apbs (apbs*);;All Files (*)')[0]
        if fnames:
            form.apbs_exe.setText(fnames[0])

    @form.pdb2pqr_exe_browse.clicked.connect
    def _():
        fnames = getOpenFileNames(None, filter='pdb2pqr (pdb2pqr*);;All Files (*)')[0]
        if fnames:
            form.pdb2pqr_exe.setText(fnames[0])

    # hook up events
    form.input_sele.currentIndexChanged.connect(
        lambda: check_sele_timer.start(0))
    form.input_sele.editTextChanged.connect(
        lambda: check_sele_timer.start(1000))

    form.button_ok.clicked.connect(run)

    # "Register" opens a web browser
    @form.button_register.clicked.connect
    def _():
        import webbrowser
        webbrowser.open("http://www.poissonboltzmann.org/")

    @form.button_load.clicked.connect
    def _():
        fnames = getOpenFileNames(
            None, filter='APBS Input (*.in);;All Files (*)')[0]
        if fnames:
            contents = load_apbs_in(form, fnames[0])
            set_apbs_in(contents)

    @form.button_reset.clicked.connect
    def _():
        set_apbs_in(electrostatics.template_apbs_in)

    form._dialog.show()
    form._dialog.resize(500, 600)


def load_apbs_in(form, filename, contents=''):
    import shlex
    from pymol import cmd, importing

    wdir = os.path.dirname(filename)

    if not contents:
        contents = cmd.file_read(filename)

    if not isinstance(contents, str):
        contents = contents.decode()

    sectionkeys = ('read', 'elec', 'apolar', 'print')
    section = ''
    insert_write_pot = True

    lines = []

    for line in contents.splitlines():
        a = shlex.split(line)
        key = a[0].lower() if a else ''

        if not section:
            if key in sectionkeys:
                section = key
        elif key == 'end':
            if section == 'elec' and insert_write_pot:
                lines.append('write pot dx "{mapfile}"')
            section = ''
        elif section == 'read':
            if len(a) > 2 and key in ('charge', 'kappa', 'mol', 'parm', 'pot'):
                filename = os.path.join(wdir, a[2])
                if os.path.exists(filename):
                    format = a[1].lower()
                    if key == 'mol' and format in ('pqr', 'pdb'):
                        # load into PyMOL and update selection dropdown
                        oname = importing.filename_to_objectname(a[2])
                        oname = cmd.get_unused_name(oname, 0)
                        cmd.load(filename, oname, format=format)
                        form.input_sele.addItem(oname)
                        form.input_sele.setEditText(oname)

                    # absolute path in input file
                    a[2] = '"' + filename + '"'
                    line = ' '.join(a)
                else:
                    QMessageBox.warning(
                        form._dialog, "Warning",
                        f'Warning: File "{filename}" does not exist')

        elif section == 'elec':
            if key == 'write':
                if a[1:4] == ['pot', 'dx', "{mapfile}"]:
                    insert_write_pot = False

        lines.append(line)

    return '\n'.join(lines)


def __init_plugin__(app=None):
    from pymol.plugins import addmenuitemqt as addmenuitem
    addmenuitem('APBS Electrostatics', dialog)



import sys
import os
import threading
import json

_SHORTCUTS_SAVE_FILE = u'~/.pymol/shortcuts_save.json'

from os.path import expanduser, expandvars

def get_shortcut_save_filename():
    '''
    Returns the users shortcuts_save filename.
    '''
    filename = expandvars(_SHORTCUTS_SAVE_FILE)
    filename = expanduser(filename)
    return filename

def save_shortcuts(shortcuts_dict):
    '''
    Saves the contents of shortcuts_dict to the filename supplied by
    get_shortcut_save_filename().
    '''
    save_file = get_shortcut_save_filename()

    try:
        os.makedirs(os.path.dirname(save_file), 0o750)
    except OSError:
        # This will trigger if the directory has already been made for license.
        pass

    try:
        with open(save_file, 'w') as savefile:
            json.dump(shortcuts_dict, savefile)
            print("Saved shortcuts to file ", save_file)
    except Exception as e:
        print("Unable to save to file.")

def load_shortcuts_dict():
    '''
    Attempts to load the filename from get_shortcut_save_filename.
    Quietly passes file doesn't exist (i.e. nothing has been saved yet)
    Dictionary is returned through save_dict.
    '''
    save_dict = {}
    save_file = get_shortcut_save_filename()
    try:
        with open(save_file) as save_dict_file:
            save_dict = json.load(save_dict_file)
    except FileNotFoundError as file_error:
        # This represents the case where no save file has been created yet and should pass quietly. 
        pass
    except Exception as e:
        print("No shortcut save file has been loaded.")
        print(e)
    return save_dict

def setkey_from_dict(save_dict,cmd):
    '''
    Iterates over a dictionary and sets the corresponding keys with cmd.set_key.
    '''
    for key in save_dict:
        if save_dict[key][2]:
            cmd.set_key(key,save_dict[key][2])

def load_and_set(cmd):
    '''
    Loads save_dict from file, sets using setkey_from_dict. Returns save_dict.
    '''
    save_dict = load_shortcuts_dict()
    setkey_from_dict(save_dict,cmd)
    return save_dict
